﻿// Fill out your copyright notice in the Description page of Project Settings.


#include "GOAPAgent.h"

#include "Actions/Combo_P.h"
#include "Actions/DefaultAttack.h"


// Sets default values for this component's properties
UGOAPAgent::UGOAPAgent()
{
	// Set this component to be initialized when the game starts, and to be ticked every frame.  You can turn these features
	// off to improve performance if you don't need them.
	PrimaryComponentTick.bCanEverTick = true;

	// ...
}


// Called when the game starts
void UGOAPAgent::BeginPlay()
{
	Super::BeginPlay();

	CombatSystem = Cast<ATurnBaseCombatV2>(GetWorld()->GetGameState());

	UGOAPAction* DefaultAttack = NewObject<UGOAPAction>(GetWorld(), UDefaultAttack::StaticClass());
	DefaultAttack->Init();
	AvailableActions.Add(DefaultAttack);
	UGOAPAction* Combo_P = NewObject<UGOAPAction>(GetWorld(), UCombo_P::StaticClass());
	Combo_P->Init();
	AvailableActions.Add(Combo_P);
}


TMap<FString, int> UGOAPAgent::GetWorldState()
{
	TMap<FString, int> WorldState;
	WorldState.Add("PlayerHealth", *CombatSystem->PlayerHealth);
	WorldState.Add("EnemyHealth", *CombatSystem->EnemyHealth);
	WorldState.Add("ProbertiumResource", CombatSystem->EnemyProbertiumResource);
	WorldState.Add("EisResource", CombatSystem->EnemyEisResource);
	WorldState.Add("AzosResource", CombatSystem->EnemyAzosResource);
	WorldState.Add("IroquoidResource", CombatSystem->EnemyIroquoidResource);
	return WorldState;
}

bool UGOAPAgent::BuildActionGraph(UActionNode* Parent, TArray<UActionNode*>& SuccessfulLeaves, TMap<FString, int> WorldState, TMap<FString, int> AgentGoals)
{
	bool bFoundAPlan = false;

	for (UGOAPAction* Action : AvailableActions)
	{
		if (Action->CheckPreConditions(WorldState))
		{
			TMap<FString, int> NewWorldState = Action->ApplyEffects(WorldState);
			UActionNode* NewNode = NewObject<UActionNode>();
			NewNode->Init(Parent, Action, NewWorldState, Parent->Cost + Action->ActionCost);

			if (NewWorldState["PlayerHealth"] <= 0)
			{
				SuccessfulLeaves.Add(NewNode);
				bFoundAPlan = true;
			}
			else
			{
				TArray<UActionNode*> NewLeaves;
				if (BuildActionGraph(NewNode, NewLeaves, NewWorldState, AgentGoals))
				{
					SuccessfulLeaves.Append(NewLeaves);
					bFoundAPlan = true;
				}
			}
		}
	}

	return bFoundAPlan;
}

// Called every frame
void UGOAPAgent::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	// ...
}

TArray<UGOAPAction*> UGOAPAgent::Plan(TMap<FString, int> WorldState, TMap<FString, int> AgentGoals)
{
	TArray<UGOAPAction*> ActionPlan;
	UActionNode* RootActionNode = NewObject<UActionNode>();
	RootActionNode->Init(nullptr, nullptr, WorldState, 0);
	TArray<UActionNode*> SuccessfulLeaves;

	if (BuildActionGraph(RootActionNode, SuccessfulLeaves, WorldState, AgentGoals))
	{
		//Find the cheapest leaf
		float LowestRunningCost = 9999999;
		UActionNode* CheapestLeaf = nullptr;
		for (UActionNode* Leaf : SuccessfulLeaves)
		{
			if (Leaf->Cost < LowestRunningCost)
			{
				LowestRunningCost = Leaf->Cost;
				CheapestLeaf = Leaf;
			}
		}

		//Get the action plan
		TArray<UActionNode*> Result;
		UActionNode* ActionNode = CheapestLeaf;
		while (ActionNode != nullptr)
		{
			Result.Add(ActionNode);
			ActionNode = ActionNode->Parent;
		}

		//Reverse the array and only get the actions
		for (int i = Result.Num() - 1; i >= 0; i--)
		{
			ActionPlan.Add(Result[i]->Action);
		}
	}

	return ActionPlan;
}
